import json

import requests

from lowerpines.exceptions import InvalidOperationException, GroupMeApiException


class Request:
    def __init__(self, gmi):
        self.gmi = gmi
        self.result = self.parse(self.execute())

    base_url = "https://api.groupme.com/v3"

    def url(self):
        raise NotImplementedError

    def mode(self):
        raise NotImplementedError

    def parse(self, response):
        raise NotImplementedError

    def args(self):
        return {}

    def execute(self):
        params = {}
        headers = {'X-Access-Token': self.gmi.access_token, 'User-Agent': 'GroupYouLibrary/1.0'}
        if self.mode() == "GET":
            params.update(self.args())
            r = requests.get(url=self.url(), params=params, headers=headers)
        elif self.mode() == "POST":
            headers['Content-Type'] = 'application/json'
            r = requests.post(url=self.url(), params=params, headers=headers, data=json.dumps(self.args()))
        elif self.mode() == "POST_RAW":
            r = requests.post(url=self.url(), params=params, headers=headers, data=self.args())
        else:
            raise InvalidOperationException()
        self.error_check(r)
        if r.content.decode('utf-8').isspace():
            return None
        else:
            return self.extract_response(r)

    def error_check(self, request):
        code = int(request.status_code)
        if 399 < code < 500:
            # noinspection PyBroadException
            try:
                text = '(JSON): ' + str(request.json()['meta']['errors'])
            except:
                text = '(TEXT): ' + str(request.text)
            raise GroupMeApiException('Something has gone wrong ' + text + ' for ' + str(self.mode()) + ' ' + str(
                self.url()) + ' with data:\n' + str(self.args()))

    def extract_response(self, response):
        return response.json()["response"]



from typing import Any

from lowerpines.gmi import GMI
from lowerpines.endpoints.request import Request


class LikeCreateRequest(Request):
    def __init__(self, gmi: GMI, conversation_id: str, message_id: str) -> None:
        self.conversation_id = conversation_id
        self.message_id = message_id
        super().__init__(gmi)

    def parse(self, response: dict) -> Any:
        return None

    def url(self) -> str:
        return self.base_url + '/messages/' + self.conversation_id + '/' + self.message_id + '/like'

    def mode(self) -> str:
        return "POST"


class LikeDestroyRequest(Request):
    def __init__(self, gmi: GMI, conversation_id: str, message_id: str) -> None:
        self.conversation_id = conversation_id
        self.message_id = message_id
        super().__init__(gmi)

    def parse(self, response: dict) -> Any:
        return None

    def url(self) -> str:
        return self.base_url + '/messages/' + self.conversation_id + '/' + self.message_id + '/unlike'

    def mode(self) -> str:
        return "POST"



from lowerpines.endpoints.object import AbstractObject, Field, RetrievableObject
from lowerpines.endpoints.request import Request
from lowerpines.exceptions import InvalidOperationException
from lowerpines.message import smart_split_complex_message


class Bot(AbstractObject, RetrievableObject):
    bot_id = Field(handler=None)
    group_id = Field()
    name = Field()
    avatar_url = Field()
    callback_url = Field()
    dm_notification = Field()

    def __init__(self, gmi, group_id=None, name=None, avatar_url=None, callback_url=None, dm_notification=None):
        self.gmi = gmi
        self.group_id = group_id
        self.name = name
        self.avatar_url = avatar_url
        self.callback_url = callback_url
        self.dm_notification = dm_notification

    @property
    def group(self):
        return self.gmi.groups.get(group_id=self.group_id)

    def save(self):
        if self.bot_id is None:
            new_data = BotCreateRequest(self.gmi, self.group_id, self.name, self.callback_url, self.avatar_url,
                                        self.dm_notification).result
            self._refresh_from_other(new_data)
        else:
            BotsUpdateRequest(self.gmi, self.bot_id, self.group_id, self.name, self.callback_url, self.avatar_url, self.dm_notification)

    def delete(self):
        if self.bot_id is None:
            raise InvalidOperationException('Cannot destroy a bot that isn\'t saved!')
        else:
            BotDestroyRequest(self.gmi, self.bot_id)

    def refresh(self):
        if self.bot_id is None:
            raise InvalidOperationException('This operation is not permitted')
        else:
            raise InvalidOperationException('This is non trivial to implement')

    def post(self, text):
        text, attachments = smart_split_complex_message(text)
        BotPostRequest(self.gmi, self.bot_id, text, attachments)

    @staticmethod
    def get_all(gmi):
        return BotIndexRequest(gmi).result

    @staticmethod
    def get(gmi, bot_id):
        pass

    def __str__(self):
        return self.name + ':' + self.group_id


class BotCreateRequest(Request):
    def __init__(self, gmi, group_id, name, callback_url=None, avatar_url=None, dm_notification=None):
        self.name = name
        self.group_id = group_id
        self.dm_notification = dm_notification
        self.callback_url = callback_url
        self.avatar_url = avatar_url
        super().__init__(gmi)

    def parse(self, response):
        return Bot.from_json(self.gmi, response['bot'])

    def mode(self):
        return "POST"

    def url(self):
        return self.base_url + "/bots"

    def args(self):
        post_dict = {
            'bot': {
                'name': self.name,
                'group_id': self.group_id,
            }
        }
        if self.avatar_url is not None:
            post_dict['bot']['avatar_url'] = self.avatar_url
        if self.callback_url is not None:
            post_dict['bot']['callback_url'] = self.callback_url
        if self.dm_notification is not None:
            post_dict['bot']['dm_notification'] = self.dm_notification
        return post_dict


class BotPostRequest(Request):
    def __init__(self, gmi, bot_id, text, attachments=None):
        self.bot_id = bot_id
        self.text = text
        self.attachments = attachments
        super().__init__(gmi)

    def parse(self, response):
        return None

    def args(self):
        post_dict = {
            'bot_id': self.bot_id,
            'text': str(self.text),
        }
        if self.attachments is not None:
            post_dict['attachments'] = self.attachments
        return post_dict

    def mode(self):
        return "POST"

    def url(self):
        return self.base_url + '/bots/post'


class BotIndexRequest(Request):
    def parse(self, response):
        bots = []
        for bot_json in response:
            bots.append(Bot.from_json(self.gmi, bot_json))
        return bots

    def mode(self):
        return "GET"

    def url(self):
        return self.base_url + "/bots"


class BotDestroyRequest(Request):
    def __init__(self, gmi, bot_id):
        self.bot_id = bot_id
        super().__init__(gmi)

    def parse(self, response):
        return None

    def mode(self):
        return "POST"

    def url(self):
        return self.base_url + '/bots/destroy'

    def args(self):
        return {
            'bot_id': self.bot_id
        }


# --- Undocumented ---

class BotsUpdateRequest(Request):
    def __init__(self, gmi, bot_id, group_id=None, name=None, callback_url=None, avatar_url=None, dm_notification=None):
        self.group_id = group_id
        self.name = name
        self.avatar_url = avatar_url
        self.callback_url = callback_url
        self.dm_notification = dm_notification
        self.bot_id = bot_id
        super().__init__(gmi)

    def url(self):
        return self.base_url + '/bots/update'

    def mode(self):
        return "POST"

    def args(self):
        post_dict = {
            'bot': {
                'bot_id': self.bot_id,
            }
        }
        if self.group_id is not None:
            post_dict['bot']['group_id'] = self.group_id
        if self.name is not None:
            post_dict['bot']['name'] = self.name
        if self.avatar_url is not None:
            post_dict['bot']['avatar_url'] = self.avatar_url
        if self.callback_url is not None:
            post_dict['bot']['callback_url'] = self.callback_url
        if self.dm_notification is not None:
            post_dict['bot']['dm_notification'] = self.dm_notification
        if self.bot_id is not None:
            post_dict['bot']['bot_id'] = self.bot_id
        return post_dict

    def parse(self, response):
        pass



from lowerpines.manager import AbstractManager
from lowerpines.endpoints.bot import Bot


class BotManager(AbstractManager):
    def create(self, group, name, callback_url=None, avatar_url=None):
        bot = Bot(self.gmi, group.group_id, name, avatar_url, callback_url)
        bot.save()
        return bot

    def _all(self):
        return Bot.get_all(self.gmi)



from unittest import TestCase

from lowerpines.exceptions import NoneFoundException, MultipleFoundException

from lowerpines.manager import AbstractManager
from lowerpines.gmi import GMI


class MockManager(AbstractManager):
    def _all(self):
        return [0, 1, 2, 3, 4]


class MockType:
    def __init__(self, val):
        self.val = val


def mm_instance(content=None):
    return MockManager(GMI('access_token_here'), content)


class TestAbstractManager(TestCase):
    def setUp(self):
        self.mm = mm_instance()
        self.mt = MockType(3)
        self.cmm = mm_instance([self.mt])
        self.cmm_multiple = mm_instance([MockType(14), MockType(14)])

    def test_all_case(self):
        for i, m in enumerate(self.mm):
            self.assertEqual(i, m)

    def test_custom_case(self):
        arr = [2, 5, 12, 77]
        mm = mm_instance(arr)
        for a, m in zip(arr, mm):
            self.assertEqual(a, m)

    def test_lazy_len_call(self):
        self.assertEqual(len(self.mm), 5)

    def test_lazy_get_item_call(self):
        self.assertEqual(self.mm[3], 3)

    def test_get_item(self):
        self.assertEqual(self.cmm.get(val=3), self.mt)

    def test_get_item_none_found(self):
        with self.assertRaises(NoneFoundException):
            self.cmm.get(val=5)

    def test_get_item_multiple_found(self):
        with self.assertRaises(MultipleFoundException):
            self.cmm_multiple.get(val=14)

    def test_filter_single(self):
        self.assertEqual(self.cmm.filter(val=3)._content, [self.mt])

    def test_filter_multiple(self):
        self.assertEqual(self.cmm_multiple.filter(val=14)._content, [self.cmm_multiple[0], self.cmm_multiple[1]])




from lowerpines.endpoints.object import AbstractObject, Field
from lowerpines.endpoints.request import Request


class Block(AbstractObject):
    user_id = Field()
    blocked_user_id = Field()
    created_at = Field()

    def __init__(self, gmi):
        self.gmi = gmi

    def get_all(self, user_id):
        return BlockIndexRequest(self.gmi, user_id).result

    @staticmethod
    def block_exists(gmi, user_id, other_user_id):
        return BlockBetweenRequest(gmi, user_id, other_user_id).result

    @staticmethod
    def block(gmi, user_id, other_user_id):
        BlockCreateRequest(gmi, user_id, other_user_id)

    @staticmethod
    def unblock(gmi, user_id, other_user_id):
        BlockUnblockRequest(gmi, user_id, other_user_id)


class BlockIndexRequest(Request):
    def __init__(self, gmi, user_id):
        self.user_id = user_id
        super().__init__(gmi)

    def mode(self):
        return "GET"

    def parse(self, response):
        blocks = list()
        for block_json in response['blocks']:
            blocks.append(Block.from_json(self.gmi, block_json))
        return blocks

    def args(self):
        return {
            'user': self.user_id
        }

    def url(self):
        return self.base_url + '/blocks'


class BlockBetweenRequest(Request):
    def __init__(self, gmi, user_id, other_user_id):
        self.user_id = user_id
        self.other_user_id = other_user_id
        super().__init__(gmi)

    def mode(self):
        return "GET"

    def parse(self, response):
        return response['between']

    def args(self):
        return {
            'user': self.user_id,
            'otherUser': self.other_user_id
        }

    def url(self):
        return self.base_url + '/blocks/between'


class BlockCreateRequest(Request):
    def __init__(self, gmi, user_id, other_user_id):
        self.user_id = user_id
        self.other_user_id = other_user_id
        super().__init__(gmi)

    def mode(self) -> str:
        return "POST"

    def parse(self, response):
        return None

    def args(self):
        return {
            'user': self.user_id,
            'otherUser': self.other_user_id
        }

    def url(self):
        return self.base_url + '/blocks'


class BlockUnblockRequest(Request):
    def __init__(self, gmi, user_id, other_user_id):
        self.user_id = user_id
        self.other_user_id = other_user_id
        super().__init__(gmi)

    def mode(self):
        return "POST"

    def parse(self, response):
        return None

    def args(self):
        return {
            'user': self.user_id,
            'otherUser': self.other_user_id
        }

    def url(self):
        return self.base_url + '/blocks/delete'



class InvalidOperationException(Exception):
    pass


class NoneFoundException(Exception):
    pass


class MultipleFoundException(Exception):
    pass


class GroupMeApiException(Exception):
    pass



import requests

_gmi_objects = []


def get_gmi(access_token):
    for gmi in _gmi_objects:
        if gmi.access_token == access_token:
            return gmi
    gmi = GMI(access_token)
    _gmi_objects.append(gmi)
    return gmi


class GMI:
    def __init__(self, access_token):
        self.access_token = access_token

        self.groups = None
        self.bots = None
        self.chats = None
        self.user = None

        self.refresh()

    def refresh(self):
        from lowerpines.group import GroupManager
        from lowerpines.bot import BotManager
        from lowerpines.chat import ChatManager
        from lowerpines.user import UserManager

        self.groups = GroupManager(self)
        self.bots = BotManager(self)
        self.chats = ChatManager(self)
        self.user = UserManager(self)

    def convert_image_url(self, url):
        from lowerpines.endpoints.image import ImageConvertRequest
        return ImageConvertRequest(self, requests.get(url).content).result




class Field:
    def __init__(self, handler=str, api_name=None):
        self.name = ''
        self.handler = handler
        self.api_name = api_name

    def set_name(self, name):
        self.name = name
        if self.api_name is None:
            self.api_name = name


class AbstractObjectType(type):
    def __new__(mcs, name, bases, attrs):
        new_attrs = {}
        fields = []
        for attr_name, attr_value in attrs.items():
            if type(attr_value) == Field:
                attr_value.set_name(attr_name)
                fields.append(attr_value)
                if attr_value.handler is None:
                    new_attrs[attr_name] = None
                else:
                    new_attrs[attr_name] = attr_value.handler()
            else:
                new_attrs[attr_name] = attr_value
        new_attrs['_fields'] = fields

        return super(AbstractObjectType, mcs).__new__(mcs, name, bases, new_attrs)


class AbstractObject(metaclass=AbstractObjectType):
    _fields = []

    def _refresh_from_other(self, other):
        if other is not None:
            for field in self._fields:
                setattr(self, field.name, getattr(other, field.name))
        self.on_fields_loaded()

    def on_fields_loaded(self):
        pass

    @classmethod
    def from_json(cls, gmi, json_dict, *args):
        obj = cls(gmi, *args)

        for field in obj._fields:
            json_val = json_dict
            for val in field.api_name.split('.'):
                json_val = json_val.get(val, None)
            if field.handler is not None:
                handled_val = field.handler(json_val)
            else:
                handled_val = json_val
            setattr(obj, field.name, handled_val)
        obj.on_fields_loaded()
        return obj


class RetrievableObject:
    def save(self):
        raise NotImplementedError

    def refresh(self):
        raise NotImplementedError

    @staticmethod
    def get(gmi, *args):
        raise NotImplementedError



from lowerpines.manager import AbstractManager
from lowerpines.endpoints.group import Group


class GroupManager(AbstractManager):
    def _all(self):
        return Group.get_all(self.gmi)

    def former(self):
        return GroupManager(self.gmi, Group.get_former(self.gmi))

    def join(self, group_id, share_token):
        return Group.join(self.gmi, group_id, share_token)

    def rejoin(self, group_id):
        return Group.rejoin(self.gmi, group_id)



from lowerpines.endpoints.request import Request
from lowerpines.endpoints.message import Message


class Leaderboard:
    def __init__(self, gmi, group_id):
        self.gmi = gmi
        self.group_id = group_id

    def _for_period(self, period):
        return LeaderboardIndexRequest(self.gmi, self.group_id, period).result

    def for_today(self):
        return self._for_period('day')

    def for_week(self):
        return self._for_period('week')

    def for_month(self):
        return self._for_period('month')

    def my_likes(self):
        return LeaderboardMyLikesRequest(self.gmi, self.group_id).result

    def my_hits(self):
        return LeaderboardMyHitsRequest(self.gmi, self.group_id).result


class LeaderboardIndexRequest(Request):
    def __init__(self, gmi, group_id, period):
        self.group_id = group_id
        if period not in ['day', 'week', 'month']:
            raise ValueError('Period must be one of: day, week, or month')
        self.period = period
        super().__init__(gmi)

    def parse(self, response):
        messages = []
        for message_json in response['messages']:
            messages.append(Message.from_json(self.gmi, message_json))
        return messages

    def url(self):
        return self.base_url + '/groups/' + self.group_id + '/likes'

    def mode(self):
        return "GET"


class LeaderboardMyLikesRequest(Request):
    def __init__(self, gmi, group_id):
        self.group_id = group_id
        super().__init__(gmi)

    def parse(self, response):
        messages = []
        for message_json in response['messages']:
            messages.append(Message.from_json(self.gmi, message_json))
        return messages

    def url(self):
        return self.base_url + '/groups/' + self.group_id + '/likes/mine'

    def mode(self):
        return "GET"


class LeaderboardMyHitsRequest(Request):
    def __init__(self, gmi, group_id):
        self.group_id = group_id
        super().__init__(gmi)

    def parse(self, response):
        messages = []
        for message_json in response['messages']:
            messages.append(Message.from_json(self.gmi, message_json))
        return messages

    def url(self):
        return self.base_url + '/groups/' + self.group_id + '/likes/for_me'

    def mode(self):
        return "GET"
